#define maxRows 240 // 24
#define maxCols 500 // 50

int rowAndColWithinBounds(int row, int col)
{
	return row >= 0 && row < maxRows && col >= 0 && col < maxCols;
}

void addAnt(struct AntList *list, int id, int row, int col, int rank)
{
	struct Ant newAnt;
	newAnt.AntId = MaxAnts*rank + id;
	newAnt.AtPos.Row = row;
	newAnt.AtPos.Col = col;
	newAnt.MovePos.Row = 0;
	newAnt.MovePos.Col = 0;
	newAnt.HasResource = 0;
	newAnt.IsFollowingFlavor = 0;
	
	int percent = rand() % 100;
	if(percent < 80)
		newAnt.Caste = 0;
	else
		newAnt.Caste = 1;
	if(newAnt.Caste == 0)		
		newAnt.ViewDist = 2;
	else if(newAnt.Caste == 1) // Spher
		newAnt.ViewDist = 4;
	
	appendAnt(list, newAnt);
}

inline struct Vector2 getCoordinate(int fromRow, int fromCol, int toRow, int toCol)
{
	struct Vector2 coord;
	coord.Row = toRow - fromRow;
	coord.Col = toCol - fromCol;
	coord.IsEmpty = 0;
	return coord;
}

/*struct AntList* getAntsOnSamePostion(struct Ant *a)
{
	struct AntList *antList = mallocAntList();
	int i;
	for(i=0; i < AntsOut; i++)
	{
		struct Ant curAnt = Ants[i];
		if(curAnt.AntId != a->AntId && curAnt.AtPos.Row == a->AtPos.Row && curAnt.AtPos.Col == a->AtPos.Col)
		{
			struct AntList *newElement = mallocAntList();
			newElement->Value = curAnt;
			antList->Next = newElement;
		}
	}
	return antList;
}*/


inline int antOnResource(struct VectorList *resourceList, int row, int col)
{
	struct VectorList *currentResource = resourceList;
	while(currentResource->Next != NULL)
	{
		currentResource = currentResource->Next;
		struct Vector2 sugarPos = currentResource->Value;
		if(sugarPos.Row == row && sugarPos.Col == col)
			return 1;		
	}
	return 0;
}

void giveDirections(struct Ant *a, struct Vector2 vec)
{
	a->MovePos.Row = vec.Row;
	a->MovePos.Col = vec.Col;
}


inline struct Vector2 antSeesRessource(struct VectorList *resourceList, int atRow, int atCol, int dist)
{
	struct Vector2 vec;
	vec.IsEmpty = 1;
	int minDistance = maxRows + maxCols;
	
	struct VectorList *currentResource = resourceList;
	while(currentResource->Next != NULL)
	{
		currentResource = currentResource->Next;
		
		//struct Vector2 vectorToRes = getCoordinate(atRow, atCol, currentResource->Value.Row, currentResource->Value.Col);
		//int distanceToResource = abs(vectorToRes.Row) + abs(vectorToRes.Col);
		int distanceToResource = abs(currentResource->Value.Row - atRow) + abs(currentResource->Value.Col - atCol);
		if(distanceToResource <= dist && distanceToResource < minDistance)
		{
			vec.IsEmpty = 0;
			vec.Row = currentResource->Value.Row;
			vec.Col = currentResource->Value.Col;
			minDistance = distanceToResource;
		}
	}
	
	return vec;
}

inline struct Vector2 antSeesFlavor(struct ListFlavor *flavorList, int atRow, int atCol, int dist)
{
	struct Vector2 vec;
	vec.IsEmpty = 1;
	int maxDistance = 0;
	
	struct ListFlavor *currentFlavor = flavorList;
	while(currentFlavor->Next != NULL)
	{
		currentFlavor = currentFlavor->Next;
		
		//struct Vector2 vectorToFlavor = getCoordinate(atRow, atCol, currentFlavor->Value.Row, currentFlavor->Value.Col);
		//int distanceToFlavor = abs(vectorToFlavor.Row) + abs(vectorToFlavor.Col);
		int distanceToFlavor = abs(currentFlavor->Value.Row - atRow) + abs(currentFlavor->Value.Col - atCol);
		
		//struct Vector2 vectorToBau = getCoordinate(currentFlavor->Value.Row, currentFlavor->Value.Col, Bau.AtRow, Bau.AtCol);
		//int distanceFromBau = abs(vectorToBau.Row) + abs(vectorToBau.Col); 
		int distanceFromBau = abs(Bau.AtRow - currentFlavor->Value.Row) + abs(Bau.AtCol - currentFlavor->Value.Col); 
			
		if(distanceToFlavor <= dist && distanceFromBau > maxDistance)
		{
			vec.IsEmpty = 0;
			vec.Row = currentFlavor->Value.Row;
			vec.Col = currentFlavor->Value.Col;
			maxDistance = distanceFromBau;
			//printf("distanceToFlavor %d  distanceFromBau %d\n", distanceToFlavor, distanceFromBau);
		}
	}

	return vec;
}

void removeFlavor(struct ListFlavor *flavorList, int row, int col)
{
	struct ListFlavor *currentFlavor = flavorList;
	while(currentFlavor->Next != NULL)
	{	
		struct ListFlavor *lastFlavor = currentFlavor;
		currentFlavor = currentFlavor->Next;
		if(currentFlavor->Value.Row == row && currentFlavor->Value.Col == col)
			lastFlavor->Next = currentFlavor->Next;
	}
}


// 1 if resource could be gathered, 0 else
int gatherResource(struct VectorList *resourceList, struct Vector2 *curVector)
{
	struct VectorList *currentResource = resourceList;
	while(currentResource->Next != NULL)
	{
		struct VectorList *lastResource = currentResource;
		currentResource = currentResource->Next;
		if(curVector->Row == currentResource->Value.Row && curVector->Col == currentResource->Value.Col)
		{						
			currentResource->Value.Resources--;
			if(currentResource->Value.Resources == 0)
			{
				// printf("MPI_ResourceDepleted at %d %d\n", currentResource->Value.Row, currentResource->Value.Col);
				lastResource->Next = currentResource->Next;
			}
			return 1;
		}
		//printf("cur resource entry %d\n", currentResource->Value.Resources);
	}
	return 0;
}

void removeResource(struct VectorList *resourceList, int row, int col)
{
	struct VectorList *currentResource = resourceList;
	while(currentResource->Next != NULL)
	{
		struct VectorList *lastResource = currentResource;
		currentResource = currentResource->Next;
		if(row == currentResource->Value.Row && col == currentResource->Value.Col)
		{
			// printf("delete resource for worker at ");
			lastResource->Next = currentResource->Next;
			return;
		}
		//printf("cur resource entry %d\n", currentResource->Value.Resources);
	}
}


