// module load  mpich2

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define	consoleOutput 0
#define fileOutput 0
#define debugOutput 0

#define emptyField 0
#define antField 1
#define flavorField 80
#define resourceField 90
#define bauField 100

#define maxRows 240
#define maxCols 500
#define MaxAnts 1000
#define SimulationLength 10000
#define ResourcesPerField 5 // defines how much sugar is on one resource field

#define	TASKS 5
#define MPI_CreateAnt 100
#define MPI_MoveAnt 200
#define MPI_IsAntOnResource 300
#define MPI_AntSeesResource 301
#define MPI_AntSeesFlavor 302
#define MPI_AddFlavor 700
#define MPI_RemoveFlavor 701
#define MPI_IncrementResource 702
#define MPI_ResourcePosition 703
#define MPI_ResourceGathered 704
#define MPI_ResourceDepleted 705

#define MPI_MovementDone 800
#define MPI_SlaveReady 801

/*
Sequentiell:
Laufzeiten Linux mit consoleOutput = false, 
Sekunden /  maxRows / maxCols  / MaxAnts / SimulationLength 
55		 /  240	 / 500		/ 30.000	 / 15.000


Parallel:
Laufzeiten Linux mit consoleOutput = false, 
Sekunden / maxRows / maxCols  / MaxAnts / SimulationLength / Threads

11				 240 	   500		  500		1.000			 / 3
7				 240 	   500		  500		1.000			 / 4
8				 240 	   500		  500		1.000			 / 5

Nach append or update flavor optimierung:

9				 240 	   500		  500		1.000			 / 3
6				 240 	   500		  500		1.000			 / 4
6				 240 	   500		  500		1.000			 / 5
6				 240 	   500		  500		1.000			 / 10

*/


struct Vector2
{
	int Row;
	int Col;
	int IsEmpty;
	int Resources;
};

struct Ant 
{
	int AntId; // each and has a unique number
	struct Vector2 AtPos;	// current pos
	struct Vector2 MovePos; // move to pos
	int HasResource;
	int IsAlive;
	int ViewDist;
	int Caste;
};

struct Flavor // defines a "Markierung"
{
	int Row;
	int Col;
	int IsEmpty;
	int Duration;
};

struct AntsBau
{
	int AtRow;	// row pos.
	int AtCol;  // column pos.
	int GatheredResources;
};

struct ListV2 // a Vector2 list
{
	struct Vector2 Value;
	struct ListV2 *Next;
};

struct ListFlavor
{
	struct Flavor Value;
	struct ListFlavor *Next;
};

struct ListAnt
{
	struct Ant Value;
	struct ListAnt *Next;
};

// global vars
struct AntsBau Bau; // initialized for all threads
/* map is used in thread 0 for debugOutput, 
in thread 1 to check if objects are in sight
the other thread do not use this global var
*/

struct ListAnt* mallocAntList()
{
	struct ListAnt *newElement = (struct ListAnt*) malloc(sizeof(struct ListAnt));
	newElement->Next = NULL;
	return newElement;
}

struct ListV2* mallocVector2List()
{
	struct ListV2 *newElement = (struct ListV2*) malloc(sizeof(struct ListV2));
	newElement->Next = NULL;
	return newElement;
}

struct ListFlavor* mallocFlavorList()
{
	struct ListFlavor *newElement = (struct ListFlavor*) malloc(sizeof(struct ListFlavor));
	newElement->Next = NULL;
	return newElement;
}

void appendVector2(struct ListV2 *list, struct Vector2 data)
{
	struct ListV2 *currentResource = list;
	while(currentResource->Next != NULL)
	{
		currentResource = currentResource->Next;
	}
	struct ListV2 *newElement = mallocVector2List();		
	newElement->Value = data;

	currentResource->Next = newElement;
}

void appendFlavor(struct ListFlavor *list, struct Flavor data)
{
	struct ListFlavor *curFlav = list;
	while(curFlav->Next != NULL)
	{
		curFlav = curFlav->Next;
		if(curFlav->Value.Row == data.Row && curFlav->Value.Col == data.Col)
		{
			// flavor on that pos. already exists. update the duration instead of creating a new entry
			curFlav->Value.Duration = data.Duration;
			return;
		}
	}
	struct ListFlavor *newElement = mallocFlavorList();		
	newElement->Value = data;

	curFlav->Next = newElement;
}

void appendAnt(struct ListAnt *list, struct Ant data)
{
	struct ListAnt *currentAnt = list;
	while(currentAnt->Next != NULL)
	{
		currentAnt = currentAnt->Next;
	}
	struct ListAnt *newElement = mallocAntList();		
	newElement->Value = data;
	
	currentAnt->Next = newElement;
}

void freeVector2(struct ListV2 *list)
{
	struct ListV2 *currentVector = list;
	while(currentVector->Next != NULL)
	{	
		struct ListV2 *lastResource = currentVector;
		currentVector = currentVector->Next;
		free(lastResource);
	}
	free(currentVector);
}

void freeFlavor(struct ListFlavor *list)
{
	struct ListFlavor *currentResource = list;
	while(currentResource->Next != NULL)
	{	
		struct ListFlavor *lastResource = currentResource;
		currentResource = currentResource->Next;
		free(lastResource);
	}
	free(currentResource);
}

void freeAnt(struct ListAnt *list)
{
	struct ListAnt *currentResource = list;
	while(currentResource->Next != NULL)
	{	
		struct ListAnt *lastResource = currentResource;
		currentResource = currentResource->Next;
		free(lastResource);
	}
	free(currentResource);
}

void addAnt(struct ListAnt *list, int id, int row, int col)
{
	struct Ant newAnt;
	newAnt.AntId = id;
	newAnt.AtPos.Row = row;
	newAnt.AtPos.Col = col;
	newAnt.MovePos.Row = 0;
	newAnt.MovePos.Col = 0;
	newAnt.HasResource = 0;

	int percent = rand() % 100;
	if(percent < 80)
		newAnt.Caste = 0;
	else
		newAnt.Caste = 1;
	if(newAnt.Caste == 0)		
		newAnt.ViewDist = 2;
	else if(newAnt.Caste == 1) // Späher
		newAnt.ViewDist = 4;
	
	appendAnt(list, newAnt);
}

struct Vector2 getCoordinate(int fromRow, int fromCol, int toRow, int toCol)
{
	struct Vector2 coord;
	coord.Row = toRow - fromRow;
	coord.Col = toCol - fromCol;
	coord.IsEmpty = 0;
	return coord;
}

/*struct ListAnt* getAntsOnSamePostion(struct Ant *a)
{
	struct ListAnt *antList = mallocAntList();
	int i;
	for(i=0; i < AntsOut; i++)
	{
		struct Ant curAnt = Ants[i];
		if(curAnt.AntId != a->AntId && curAnt.AtPos.Row == a->AtPos.Row && curAnt.AtPos.Col == a->AtPos.Col)
		{
			struct ListAnt *newElement = mallocAntList();
			newElement->Value = curAnt;
			antList->Next = newElement;
		}
	}
	return antList;
}*/

int antOnResource(struct ListV2 *resourceList, int row, int col)
{
	struct ListV2 *currentResource = resourceList;
	while(currentResource->Next != NULL)
	{
		currentResource = currentResource->Next;
		struct Vector2 sugarPos = currentResource->Value;
		if(sugarPos.Row == row && sugarPos.Col == col)
			return 1;		
	}
	return 0;
}

void giveDirections(struct Ant *a, struct Vector2 vec)
{
	a->MovePos.Row = vec.Row;
	a->MovePos.Col = vec.Col;
}

void giveDirectionsToBau(struct Ant *a)
{
	struct Vector2 vec = getCoordinate(a->AtPos.Row, a->AtPos.Col, Bau.AtRow, Bau.AtCol);
	giveDirections(a, vec);
}

int antIsAtBau(struct Ant a)
{
	return a.AtPos.Row == Bau.AtRow && a.AtPos.Col == Bau.AtCol;
}

int rowAndColWithinBounds(int row, int col)
{
	return row >= 0 && row < maxRows && col >= 0 && col < maxCols;
}

void initBau()
{
	Bau.AtRow = maxRows / 2;
	Bau.AtCol = maxCols / 2;
	Bau.GatheredResources = 0;
}

void initResource(struct ListV2 *resourceList)
{	
	// determine the sugar positions
	int sugarsToCreate = maxRows * maxCols / 300; // exactly 4 sugar fields for 24 rows * 50 cols = 1200 fields	
	int row, col;
	srand (time(NULL));
	for(sugarsToCreate; sugarsToCreate > 0; sugarsToCreate--)
	{		
		int distanceFromBau;
		struct Vector2 sugarPos;
		do
		{
			sugarPos.Row = rand() % maxRows;
			sugarPos.Col = rand() % maxCols;
			struct Vector2 vectorToBau = getCoordinate(sugarPos.Row, sugarPos.Col, Bau.AtRow, Bau.AtCol);
			distanceFromBau = abs(vectorToBau.Row) + abs(vectorToBau.Col); 
		}
		while(distanceFromBau < 10);
		
		for(row = sugarPos.Row; row < sugarPos.Row + 2; row++)
			for(col = sugarPos.Col; col < sugarPos.Col + 3; col++)
			{
				struct Vector2 tmp;
				tmp.IsEmpty = 0;
				tmp.Resources = ResourcesPerField;
				tmp.Row = row;
				tmp.Col = col;
				if(rowAndColWithinBounds(row, col))
					appendVector2(resourceList, tmp);
			}
	}
}

void sendResourcesToSlaveOne(struct ListV2 *resourceList)
{
	struct ListV2 *currentResource = resourceList;
	while(currentResource->Next != NULL)
	{
		currentResource = currentResource->Next;
		int resPos[2] = { currentResource->Value.Row, currentResource->Value.Col };
		if(debugOutput)
			printf("send resource at pos %d %d to task one\n", resPos[0], resPos[1]);
		MPI_Send(resPos, 2, MPI_INT, 1, MPI_ResourcePosition, MPI_COMM_WORLD);
	}
}


inline void move(struct Vector2 *curPos, struct Vector2 *movePos)
{
	if(curPos->Row + movePos->Row >= maxRows)
		movePos->Row = maxRows - curPos->Row - 1;
	if(curPos->Col + movePos->Col >= maxCols)
		movePos->Col = maxCols - curPos->Col - 1;
	if(curPos->Row + movePos->Row < 0)
		movePos->Row = -curPos->Row;
	if(curPos->Col + movePos->Col < 0)
		movePos->Col = -curPos->Col;

	int moveRow = 0, moveCol = 0;
	if( movePos->Row != 0 || movePos->Col != 0)
	{
		if(abs(movePos->Row) > abs(movePos->Col))
			moveRow = abs(movePos->Row) / movePos->Row;
		else
			moveCol = abs(movePos->Col) / movePos->Col;
		movePos->Row -= moveRow;
		movePos->Col -= moveCol;
	}

	curPos->Row = curPos->Row+moveRow;
	curPos->Col = curPos->Col+moveCol;
}

inline void moveRnd(struct Vector2 *curPos, struct Vector2 *movePos)
{
	if(movePos->Row == 0 && movePos->Col == 0) // give new direction
	{
		movePos->Row = rand() % 11 - 5;
		movePos->Col = rand() % 11 - 5;
	}
	move(curPos, movePos); // do 1 step into direction
}

/* TODO ants in liste?!?
int antsAtPosition(struct Vector2 position, int antsOut)
{
	int i, count = 0; // i to iterate, count to increment the ant count at a position
	for(i = 0; i < antsOut; i++)
	{
		if(Ants[i].AtPos.Row == position.Row && Ants[i].AtPos.Col == position.Col)
		{
			count++;
		}
	}
	return count;
}*/

void addFlavor(struct ListFlavor *flavorList, int row,int col, int duration) {
	struct Flavor fl;
	fl.Row = row;
	fl.Col = col;
	fl.Duration = duration;
	appendFlavor(flavorList, fl);
}

void updateFlavorList(struct ListFlavor *flavorList)
{
	struct ListFlavor *currentFlavor = flavorList;
	while(currentFlavor->Next != NULL)
	{	
		struct ListFlavor *lastFlavor = currentFlavor;
		currentFlavor = currentFlavor->Next;
		if(currentFlavor->Value.Duration > 0)
			currentFlavor->Value.Duration--;
		if(currentFlavor->Value.Duration == 0)
		{
			int flavorPos[2] = { currentFlavor->Value.Row, currentFlavor->Value.Col };
			MPI_Send(flavorPos, 2, MPI_INT, 1, MPI_RemoveFlavor, MPI_COMM_WORLD);
			lastFlavor->Next = currentFlavor->Next;
		}
	}
}

void gatherResource(struct ListV2 *resourceList, struct Vector2 *curVector)
{
	struct ListV2 *currentResource = resourceList;
	while(currentResource->Next != NULL)
	{
		struct ListV2 *lastResource = currentResource;
		currentResource = currentResource->Next;
		if(curVector->Row == currentResource->Value.Row && curVector->Col == currentResource->Value.Col)
		{						
			currentResource->Value.Resources--;
			if(currentResource->Value.Resources == 0)
			{
				// printf("MPI_ResourceDepleted at %d %d\n", currentResource->Value.Row, currentResource->Value.Col);
				int resPos[2] = { currentResource->Value.Row, currentResource->Value.Col };
				MPI_Send(resPos, 2, MPI_INT, 1, MPI_ResourceDepleted, MPI_COMM_WORLD);
				lastResource->Next = currentResource->Next;
			}
			break;
		}		
		//printf("cur resource entry %d\n", currentResource->Value.Resources);
	}
}

inline void calculateAntMovement(struct Ant *a)
{
	MPI_Status status;
	int antOnResourceFlag = 0;
	int antSeesResourceFlag[2];
	int antSeesFlavorFlag[2];
	int curPos[2] = { a->AtPos.Row, a->AtPos.Col };
		
	MPI_Send(curPos, 2, MPI_INT, 0, MPI_IsAntOnResource, MPI_COMM_WORLD);
	MPI_Send(curPos, 2, MPI_INT, 1, MPI_AntSeesResource, MPI_COMM_WORLD);
	MPI_Send(curPos, 2, MPI_INT, 1, MPI_AntSeesFlavor, MPI_COMM_WORLD);
	if(debugOutput)					
		printf("Slave_Requests res + flav for pos: %d %d\n", curPos[0], curPos[1]);
		
	MPI_Probe(0, MPI_IsAntOnResource, MPI_COMM_WORLD, &status); // handled via thread 0 resource list
	MPI_Probe(1, MPI_AntSeesResource, MPI_COMM_WORLD, &status); // handled via thread 1 map
	MPI_Probe(1, MPI_AntSeesFlavor, MPI_COMM_WORLD, &status);   // handled via thread 1 map
	MPI_Recv(&antOnResourceFlag, 1, MPI_INT, 0, MPI_IsAntOnResource, MPI_COMM_WORLD, &status);
	MPI_Recv(antSeesResourceFlag, 2, MPI_INT, 1, MPI_AntSeesResource, MPI_COMM_WORLD, &status);
	MPI_Recv(antSeesFlavorFlag, 2, MPI_INT, 1, MPI_AntSeesFlavor, MPI_COMM_WORLD, &status);
	
	if(debugOutput && antOnResourceFlag)					
		printf("Slave_RequestIsAntOnResource result %d\n", antOnResourceFlag);
		
	if(a->HasResource && !antOnResourceFlag) {		
		MPI_Send(curPos, 2, MPI_INT, 0, MPI_AddFlavor, MPI_COMM_WORLD);
	}
	if(!a->HasResource && antOnResourceFlag)
	{
		MPI_Send(curPos, 2, MPI_INT, 0, MPI_ResourceGathered, MPI_COMM_WORLD);
		a->HasResource = 1;		
		giveDirectionsToBau(a);
	}	
	else if(!a->HasResource && antSeesResourceFlag[0] != -1 && antSeesResourceFlag[1] != -1)
	{
		//printf("Slave_antSeesResourceFlag && !a->HasResource\n");
		struct Vector2 movement;		
		movement = getCoordinate(a->AtPos.Row, a->AtPos.Col, antSeesResourceFlag[0], antSeesResourceFlag[1]);	
		giveDirections(a, movement);
		move(&a->AtPos, &a->MovePos);
	}		
	else if(!a->HasResource && a->Caste == 0 && antSeesFlavorFlag[0] != -1 && antSeesFlavorFlag[1] != -1) {
		//printf("Slave_antSeesFlavorFlag && !a->HasResource && a->Caste == 0\n");
		struct Vector2 movement;		
		movement = getCoordinate(a->AtPos.Row, a->AtPos.Col, antSeesFlavorFlag[0], antSeesFlavorFlag[1]);	
		giveDirections(a, movement);
		move(&a->AtPos, &a->MovePos);
	}
	else if(a->HasResource && antIsAtBau(*a))
	{
		a->HasResource = 0;
		MPI_Send(&a->AntId, 1, MPI_INT, 0, MPI_IncrementResource, MPI_COMM_WORLD);		
	}	
	else
	{
		moveRnd(&a->AtPos, &a->MovePos);
	}
	
	int newPos[2] = { a->AtPos.Row, a->AtPos.Col };
	MPI_Send(newPos, 2, MPI_INT, 0, MPI_MovementDone, MPI_COMM_WORLD);
}


MPI_Status receiveIndexOrVector(int *nbr, struct Vector2 *vec)
{
	MPI_Status status;
	MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
	
	int nelements;
	MPI_Get_count(&status, MPI_INT, &nelements);
	
	if(nelements == 1)
	{
		int antIndex;
		MPI_Recv(&antIndex, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
		*nbr = antIndex;
	}
	else if (nelements == 2)
	{
		int pos[2];
		MPI_Recv(pos, 2, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
		vec->Row = pos[0];
		vec->Col = pos[1];
		if(pos[0] != -1 && pos[1] != -1)
			vec->IsEmpty = 0;
		else
			vec->IsEmpty = 1;
	}
	else
	{
		if(debugOutput)					
			printf("nelements: %d\n", nelements);
	}
	return status;
}

void master(int ntasks)
{
	int map[maxRows][maxCols]; 
	void clearMap()
	{
		int i, j;	
		for(i = 0; i < maxRows; i++)
			for(j = 0; j < maxCols; j++)
				if(map[i][j] != emptyField)
					map[i][j] = emptyField;
	}
	
	void printMap(int antsOut)
	{
		int i, j;
		for(i = 0; i < maxRows; i++)
		{
			for(j = 0; j < maxCols; j++)
			{		
				int value = map[i][j];
				if(value == bauField)
				{
					printf("b");
				}
				else if(value == resourceField)
				{
					printf("z");
				}
				
				else if(value == emptyField)
				{
					printf(" ");
				}
				else if(value == antField)
				{
					struct Vector2 pos;
					pos.IsEmpty = 0;
					pos.Row = i;
					pos.Col = j;
					/*
					int antsAtPos = antsAtPosition(pos);
					if(antsAtPos > 9)
						printf("+");
					else
					*/
						printf("%d", antField);
				}
				else if(value == flavorField)
				{
					printf("D");
				}
				else
				{
					printf("%d", value);
				}
			}
			printf("\n");
		}
			
		/*
		int totalWorkers = 0;
		int totalScouts = 0;
		for(i = 0; i < antsOut; i++)
		{
			if(Ants[i].Caste == 0)
				totalWorkers++;
			else // == 1
				totalScouts++;
		}
		printf("wokers/scouts: %d/%d resources: %d, ", totalWorkers, totalScouts, Bau.GatheredResources);
		*/
	} // end printMap

	void initMap(struct ListV2 *resourceList, struct ListFlavor *flavorList, int currentRound)
	{
		struct ListFlavor *currentFlavor = flavorList;
		while(currentFlavor->Next != NULL)
		{
			currentFlavor = currentFlavor->Next;
			struct Flavor flavorPos = currentFlavor->Value;
			if(flavorPos.Duration > 0)
				map[flavorPos.Row][flavorPos.Col] = flavorField;			
		}
		
		int i, j;
		// draw some sugar
		struct ListV2 *currentResource = resourceList;
		while(currentResource->Next != NULL)
		{
			currentResource = currentResource->Next;
			struct Vector2 sugarPos = currentResource->Value;
			map[sugarPos.Row][sugarPos.Col] = resourceField;
		}
		
		map[Bau.AtRow][Bau.AtCol] = bauField;
	}

	time_t start;
	time(&start);

	struct ListV2 *resourceList;
	resourceList = mallocVector2List();
	initResource(resourceList);
	sendResourcesToSlaveOne(resourceList);	
	
	struct ListFlavor *flavorList;
	flavorList = mallocFlavorList();	
	
	int antsOut = 0, duration, index;
	
	for(duration = 0; duration < SimulationLength; duration++)
	{	
		int currentTask;
		if(consoleOutput)
			clearMap();
		if(antsOut < MaxAnts)
		{
			if(debugOutput)					
				printf("Round %d; Master_CreateAnt\n", duration);
			int newAntCount = MaxAnts / (SimulationLength / 2);
			if(newAntCount == 0)
				newAntCount = 1;
			
			for(newAntCount; newAntCount > 0; newAntCount--)
			{
				currentTask = (antsOut % (TASKS - 2)) + 2;
				MPI_Send(&antsOut, 1, MPI_INT, currentTask, MPI_CreateAnt, MPI_COMM_WORLD);
				// printf("create: current worker task is %d\n", currentTask);
				antsOut++;
			}
		}
		
		for(index = 0; index < antsOut; index++)
		{
			currentTask = (index % (TASKS - 2)) + 2; 
			MPI_Send(&index, 1, MPI_INT, currentTask, MPI_MoveAnt, MPI_COMM_WORLD);
			// printf("move: current worker task is %d\n", currentTask);
		}
		
		int antsDone = antsOut;
		while(antsDone != 0) // master
		{
			int antIndex = -1;
			struct Vector2 curVector;			
			MPI_Status status = receiveIndexOrVector(&antIndex, &curVector);
			
			if(debugOutput && 0)
				printf("Master_Listening\n");
		
			if(status.MPI_TAG == MPI_SlaveReady)
			{
				antsDone--;
			}
			else if(status.MPI_TAG == MPI_MovementDone)
			{
				if(consoleOutput)
					map[curVector.Row][curVector.Col] = antField;
				if(debugOutput)
				{
					if(curVector.Row < 0 || curVector.Row >= maxRows)
						printf("WARNING ant out of map: at pos %d %d\n", curVector.Row, curVector.Col);
					else if(curVector.Col < 0 || curVector.Col >= maxCols)
						printf("WARNING ant out of map: at pos %d %d\n", curVector.Row, curVector.Col);
					else
						printf("Master_NewPos %d %d\n", curVector.Row, curVector.Col);
				}
				
			}
			else if(status.MPI_TAG == MPI_IsAntOnResource)
			{
				int b = antOnResource(resourceList, curVector.Row, curVector.Col);
				if(debugOutput && b)
					printf("Master_AntOnResource Result: %d\n", b);
				MPI_Send(&b, 1, MPI_INT, status.MPI_SOURCE, MPI_IsAntOnResource, MPI_COMM_WORLD);
			}
			else if(status.MPI_TAG == MPI_AddFlavor)
			{
				struct Vector2 vectorToBau = getCoordinate(curVector.Row, curVector.Col, Bau.AtRow, Bau.AtCol);
				int distanceFromBau = abs(vectorToBau.Row) + abs(vectorToBau.Col); 
			
				addFlavor(flavorList, curVector.Row, curVector.Col, distanceFromBau*3);
				int aCurVector[2] = { curVector.Row, curVector.Col };
				MPI_Send(aCurVector, 2, MPI_INT, 1, MPI_AddFlavor, MPI_COMM_WORLD);
			}
			else if(status.MPI_TAG == MPI_ResourceGathered)
			{
				gatherResource(resourceList, &curVector);
			}			
			else if(status.MPI_TAG == MPI_IncrementResource)
			{
				Bau.GatheredResources++;
			}			
			else if(debugOutput)
			{				
				printf("error on master: %d\n", status.MPI_TAG);
				exit(1);
			}
		}
				
		if(consoleOutput)
		{
			system("clear");
			initMap(resourceList, flavorList, duration);
			printMap(antsOut);			
			printf("resources: %d, ", Bau.GatheredResources);
			printf("Round: %d/%d, ", duration+1, SimulationLength);		
		}
		
		updateFlavorList(flavorList);
		//getchar();	
	} // end for(duration < SimulationLength)

	int timeDiff = (int)time(NULL) - (int)start;
	printf("resources: %d, ", Bau.GatheredResources);
	printf("%d Ameisen in %d seconds.\n", antsOut, timeDiff);
	
	int quit = 999;
	for(index = 1; index < ntasks; index++)
	{
		MPI_Send(&quit, 1, MPI_INT, index, 666, MPI_COMM_WORLD);
	}
	
	freeVector2(resourceList);
	freeFlavor(flavorList);
	
} // end master

void resourceAndFlavorSlave()
{
	int map[maxRows][maxCols]; 
	// returns the position of a resource in the ants distance
	struct Vector2 antSeesObject(int atRow, int atCol, int dist, int objectId)
	{
		struct Vector2 vec;
		vec.IsEmpty = 1;
		int row, col, viewWidth = 0;
		int maxDistance, tempDistance;
		maxDistance=0;
		for(row = atRow - dist; row <= atRow + dist; row++)
		{
			for(col = atCol - viewWidth; col <= atCol + viewWidth; col++)
			{
				if(rowAndColWithinBounds(row, col))
				{
					if((row != atRow || col != atCol) && map[row][col] == objectId)
					{
						if(objectId == flavorField) {
							
							struct Vector2 vectorToBau = getCoordinate(Bau.AtRow, Bau.AtCol, row, col);
							tempDistance = abs(vectorToBau.Row) + abs(vectorToBau.Col); 
							
							if(tempDistance > maxDistance)
							{
								maxDistance = tempDistance;							
								//printf("new maxdistance %d\n", maxDistance);
								vec.Row = row;
								vec.Col = col;
								vec.IsEmpty = 0;
							}
						}
						else
						{				
							vec.Row = row;
							vec.Col = col;
							vec.IsEmpty = 0;
							return vec;
						}
					}
				}
			}
			if(row < atRow)
				viewWidth++;
			else
				viewWidth--;		
		}
		return vec;
	} // end antSeesObject
	

	int running = 1;
	while(running == 1) {
		
		// MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
					
		int antIndex;
		struct Vector2 curVec = { -1, -1 };
		MPI_Status status = receiveIndexOrVector(&antIndex, &curVec);
		
		if(debugOutput && 0)					
			printf("Slave_One_Listening...\n");			
				
		if(status.MPI_TAG == 666) {
			running = 0;				
			printf("Thread 1 terminiert nun!\n");
		}
		else if(status.MPI_TAG == MPI_ResourcePosition) // Resource is saved for thread 1
		{
			map[curVec.Row][curVec.Col] = resourceField;
			if(debugOutput)
				printf("Helper Thread: MPI_ResourcePosition at %d %d\n", curVec.Row, curVec.Col);
		}
		else if(status.MPI_TAG == MPI_ResourceDepleted) // Resource deleted for thread 1
		{
			map[curVec.Row][curVec.Col] = emptyField;;
			if(debugOutput)
				printf("Helper Thread: MPI_ResourceDepleted at %d %d\n", curVec.Row, curVec.Col);
		}		
		else if(status.MPI_TAG == MPI_RemoveFlavor) // Flavor deleted for thread 1
		{
			map[curVec.Row][curVec.Col] = emptyField;;
			if(debugOutput)
				printf("Helper Thread: MPI_RemoveFlavor at %d %d\n", curVec.Row, curVec.Col);
		}
		else if(status.MPI_TAG == MPI_AddFlavor)
		{
			if(map[curVec.Row][curVec.Col] != bauField && map[curVec.Row][curVec.Col] != resourceField)
				map[curVec.Row][curVec.Col] = flavorField;
			//printf("Helper Thread: MPI_AddFlavor at %d %d\n", curVec.Row, curVec.Col);
		}
		else if(status.MPI_TAG == MPI_AntSeesResource)
		{
			struct Vector2 resPos = antSeesObject(curVec.Row, curVec.Col, 2, resourceField); // row, col, dist, objectId....
			int aResPos[2] = { -1, -1 };
			if(!resPos.IsEmpty)
			{
				aResPos[0] = resPos.Row;
				aResPos[1] = resPos.Col;
			}
			MPI_Send(aResPos, 2, MPI_INT, status.MPI_SOURCE, MPI_AntSeesResource, MPI_COMM_WORLD);
			if(debugOutput && !resPos.IsEmpty)
				printf("Helper Thread: MPI_AntSeesResource:%d\n", !resPos.IsEmpty);
		}
		else if(status.MPI_TAG == MPI_AntSeesFlavor)
		{
			struct Vector2 flavPos = antSeesObject(curVec.Row, curVec.Col, 2, flavorField);
			int aflavPos[2] = { -1, -1 };
			if(!flavPos.IsEmpty)
			{
				aflavPos[0] = flavPos.Row;
				aflavPos[1] = flavPos.Col;
			}
			MPI_Send(aflavPos, 2, MPI_INT, status.MPI_SOURCE, MPI_AntSeesFlavor, MPI_COMM_WORLD);
			if(debugOutput && !flavPos.IsEmpty)
				printf("Helper Thread: MPI_AntSeesFlavor\n");
		}
		else if(debugOutput)
		{				
			printf("error on slave one: %d\n", status.MPI_TAG);
			exit(1);
		}
	}
	
	// regular end of resourceAndFlavorSlave
	
	
	
} // end of resourceAndFlavorSlave

void slave(int rank)
{	
	struct ListAnt *antList = mallocAntList();
	MPI_Status status;
	int running = 1;
	srand(time(NULL)+(rank*100)); // every slave needs its own seed for rand
	while(running == 1) {			
		MPI_Probe(0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);  //  int source, int tag, MPI_Comm comm, MPI_Status* status)
		
		if(debugOutput && 0)					
			printf("Slave_Listening...\n");			
				
		int antIndex;
		MPI_Recv(&antIndex, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);			
		
		if(status.MPI_TAG == 666) {
			running = 0;				
			printf("Thread %i terminiert nun!\n", rank);
		}		
		else if(status.MPI_TAG == MPI_CreateAnt)
		{
			if(debugOutput)					
				printf("Slave_CreateAnt id: %d\n", antIndex);
			addAnt(antList, antIndex, Bau.AtRow, Bau.AtCol);				
		}
		else if(status.MPI_TAG == MPI_MoveAnt)
		{				
			struct ListAnt *currentAnt = antList->Next;
			while(currentAnt->Next != NULL)
			{
				if(currentAnt->Value.AntId == antIndex)
					break;
				currentAnt = currentAnt->Next;
			}
			if(debugOutput)
				printf("Slave_MoveAnt id:%d\n", currentAnt->Value.AntId);
			calculateAntMovement(&(currentAnt->Value));
			MPI_Send(&antIndex, 1, MPI_INT, 0, MPI_SlaveReady, MPI_COMM_WORLD);
		}
		else if(debugOutput)
		{				
			printf("error on slave: %d\n", status.MPI_TAG);
			exit(1);
		}
	}
}

int main(argc, argv)
     int argc;
     char *argv[];
{
	initBau();
		
	int rank, ntasks;	
	MPI_Init (&argc, &argv);	
	MPI_Comm_rank (MPI_COMM_WORLD, &rank);	
	MPI_Comm_size (MPI_COMM_WORLD, &ntasks);

	if (ntasks != TASKS) {
	  printf("Must specify TASKS= %d. Terminating.\n", TASKS);
	  MPI_Finalize();
	  exit(0);
	} else if(maxRows*maxCols < 300) {
		printf("map must be at least 12x25!");
		MPI_Finalize();
		exit(0);
	}
	
	if(rank == 0) {
		master(ntasks);
	}
	else if(rank == 1) {
		resourceAndFlavorSlave();
	}	
	else {		
		slave(rank);
	}
	
	MPI_Finalize();
	return 0;
}