#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#define maxRows 24
#define maxCols 50

#define emptyField 0
#define antField 1
#define resourceField 90
#define bauField 100

#define MaxAnts 100
#define SimulationLength 100

/*Laufzeiten Windows mit consoleOutput = false, fileOutput = false:
Sekunden / RAM / maxRows / maxCols  / MaxAnts / SimulationLength / TextFileSize
55		 / 5mb / 240	 / 500		/ 30.000	 / 15.000			/ 0

Laufzeiten Windows mit consoleOutput = false, fileOutput = true:
23		 / 3mb / 240	 / 500		/ 3.000	 / 5.000				/ 160mb

*/


struct Vector2
{
	int Row;
	int Col;
	int IsEmpty;
};

struct Ant 
{
	int AntId; // each and has a unique number
	struct Vector2 AtPos;	// current pos
	struct Vector2 MovePos; // move to pos
	int HasResource;
	struct Vector2 GatheredFrom; // saves the positions from where a resource was gathered
	int IsAlive;
	int ViewDist;
	int Caste;
};

struct AntsBau
{
	int AtRow;	// row pos.
	int AtCol;  // column pos.
	int GatheredResources;
	struct Vector2 ResourcePosition;
};

struct ListV2
{
	struct Vector2 Value;
	struct ListV2 *Next;
};

struct ListAnt
{
	struct Ant Value;
	struct ListAnt *Next;
};

int map[maxRows][maxCols];

int AntsOut = 0;
struct Ant Ants[MaxAnts];

struct ListV2 *ResourceList;

struct AntsBau Bau;

const char mapFile[] = "mapFile";
FILE *f;

int consoleOutput = 1;
int fileOutput = 0;


struct ListAnt* mallocAntList()
{
	struct ListAnt *newElement = (struct ListAnt*) malloc(sizeof(struct ListAnt));
	newElement->Next = NULL;
	return newElement;
}

struct ListV2* mallocVector2List()
{
	struct ListV2 *newElement = (struct ListV2*) malloc(sizeof(struct ListV2));
	newElement->Next = NULL;
	return newElement;
}

void appendVector2(struct ListV2 *list, struct Vector2 data)
{
	struct ListV2 *currentResource = list;
	while(currentResource->Next != NULL)
	{
		currentResource = currentResource->Next;
	}
	struct ListV2 *newElement = mallocVector2List();		
	newElement->Value = data;

	currentResource->Next = newElement;
}

void appendAnt(struct ListAnt *list, struct Ant data)
{
	struct ListAnt *currentResource = list;
	while(currentResource->Next != NULL)
	{
		currentResource = currentResource->Next;
	}
	struct ListAnt *newElement = mallocAntList();
		
	newElement->Value = data;
	newElement->Next = NULL;
	currentResource->Next = newElement;
}

void freeVector2(struct ListV2 *list)
{
	struct ListV2 *currentResource = list;
	while(currentResource->Next != NULL)
	{	
		struct ListV2 *lastResource = currentResource;
		currentResource = currentResource->Next;
		free(lastResource);
	}
	free(currentResource);
}

void freeAnt(struct ListAnt *list)
{
	struct ListAnt *currentResource = list;
	while(currentResource->Next != NULL)
	{	
		struct ListAnt *lastResource = currentResource;
		currentResource = currentResource->Next;
		free(lastResource);
	}
	free(currentResource);
}


void initMap()
{
	int i, j;	
	for(i = 0; i < maxRows; i++)
	{
		for(j = 0; j < maxCols; j++)
		{
			map[i][j] = emptyField;
		}
	}
	
	for(i = 0; i < AntsOut; i++)
	{
		int row = Ants[i].AtPos.Row, col = Ants[i].AtPos.Col;
		map[row][col] = antField;
	}

	// draw some sugar
	struct ListV2 *currentResource = ResourceList;
	while(currentResource->Next != NULL)
	{
		currentResource = currentResource->Next;
		struct Vector2 sugarPos = currentResource->Value;

		for(i = sugarPos.Row; i < sugarPos.Row + 2; i++)
		{
			for(j = sugarPos.Col; j < sugarPos.Col + 3; j++)
			{
				if(map[i][j] != bauField)
					map[i][j] = resourceField;
			}
		}
	}
	map[Bau.AtRow][Bau.AtCol] = bauField;
}

void printMap()
{
	int i, j;
	initMap();

	for(i = 0; i < maxRows; i++)
	{
		for(j = 0; j < maxCols; j++)
		{		
			int value = map[i][j];
			if(value == bauField)
			{
				printf("b");
			}
			else if(value == resourceField)
			{
				printf("z");
			}
			else if(value == emptyField)
			{
				printf(" ");
			}
			else if(value == antField)
			{
				struct Vector2 pos;
				pos.IsEmpty = 0;
				pos.Row = i;
				pos.Col = j;
				int antsAtPos = antsAtPosition(pos);
				if(antsAtPos > 9)
					printf("+");
				else
					printf("%d", antsAtPos);
			}
			else
			{
				printf("%d", value);
			}
		}
		printf("\r\n");
	}
		
	int totalWorkers = 0;
	int totalScouts = 0;
	for(i = 0; i < AntsOut; i++)
	{
		if(Ants[i].Caste == 0)
			totalWorkers++;
		else // == 1
			totalScouts++;
	}
	printf("wokers/scouts: %d/%d resources: %d, ", totalWorkers, totalScouts, Bau.GatheredResources);
}

void printMapToFile()
{
	int i;
	initMap();

	// print by ants-array
	for(i = 0; i < AntsOut; i++)
	{
		int row = Ants[i].AtPos.Row, col = Ants[i].AtPos.Col;
		//if(Ants[i].IsAlive) // always false atm
		fprintf (f, "i:%d;r:%d;c:%d;z:%d;,", Ants[i].AntId, row, col, Ants[i].HasResource);
	}

	fprintf (f, "\n");
}

void addAnt(int id, int row, int col)
{
		map[row][col] = 1;

		Ants[id].AntId = id;
		Ants[id].AtPos.Row = row;
		Ants[id].AtPos.Col = col;
		Ants[id].MovePos.Row = 0;
		Ants[id].MovePos.Col = 0;
		Ants[id].GatheredFrom.IsEmpty = 1;

		int percent = rand() % 100;
		if(percent < 80)
			Ants[id].Caste = 0;
		else
			Ants[id].Caste = 1;
		if(Ants[id].Caste == 0)		
			Ants[id].ViewDist = 2;
		else if(Ants[id].Caste == 1) // Späher
			Ants[id].ViewDist = 4;
		

		AntsOut++;
}



struct ListV2* getObjectsInViewDistance(struct Ant *a, int objectId)
{
	int atRow = a->AtPos.Row, atCol = a->AtPos.Col, dist = a->ViewDist;
	struct ListV2 *headOfList = (struct ListV2*)malloc(sizeof(struct ListV2));
	headOfList->Next = NULL;
	
	int row, col, viewWidth = 0;
	for(row = atRow - dist; row <= atRow + dist; row++)
	{
		for(col = atCol - viewWidth; col <= atCol + viewWidth; col++)
		{
			if(rowAndColWithinBounds(row, col))
			{
				if((row != atRow || col != atCol) && map[row][col] == objectId)
				{
					struct Vector2 newPos;
					newPos.IsEmpty = 0;
					newPos.Row = row;
					newPos.Col = col;
					appendVector2(headOfList, newPos);
				}
			}
		}
		if(row < atRow)
			viewWidth++;
		else
			viewWidth--;		
	}
	return headOfList;
}

// returns if Ant sees a resource in its distance
// changes the move params to the destination of the resource
int antSeesObject(struct Ant *a, int objectId)
{
	int atRow = a->AtPos.Row, atCol = a->AtPos.Col, dist = a->ViewDist;
	int row, col, viewWidth = 0;
	for(row = atRow - dist; row <= atRow + dist; row++)
	{
		for(col = atCol - viewWidth; col <= atCol + viewWidth; col++)
		{
			if(rowAndColWithinBounds(row, col))
			{
				if((row != atRow || col != atCol) && map[row][col] == objectId)
				{
					return 1;
				}
			}
		}
		if(row < atRow)
			viewWidth++;
		else
			viewWidth--;		
	}
	return 0;
}

struct ListAnt* getAntsOnSamePostion(struct Ant *a)
{
	struct ListAnt *antList = mallocAntList();
	int i;
	for(i=0; i < AntsOut; i++)
	{
		struct Ant curAnt = Ants[i];
		if(curAnt.AntId != a->AntId && curAnt.AtPos.Row == a->AtPos.Row && curAnt.AtPos.Col == a->AtPos.Col)
		{
			struct ListAnt *newElement = mallocAntList();
			newElement->Value = curAnt;
			antList->Next = newElement;
		}
	}
	return antList;
}

int antOnResource(struct Ant a)
{
	return map[a.AtPos.Row][a.AtPos.Col] == resourceField;
}

void giveDirections(struct Ant *a, struct Vector2 vec)
{
	a->MovePos.Row = vec.Row;
	a->MovePos.Col = vec.Col;
}

struct Vector2 getCoordinate(int fromRow, int fromCol, int toRow, int toCol)
{
	struct Vector2 coord;
	coord.Row = toRow - fromRow;
	coord.Col = toCol - fromCol;
	coord.IsEmpty = 0;
	return coord;
}

void giveDirectionsToBau(struct Ant *a)
{
	struct Vector2 vec = getCoordinate(a->AtPos.Row, a->AtPos.Col, Bau.AtRow, Bau.AtCol);
	giveDirections(a, vec);
}

int antIsAtBau(struct Ant a)
{
	return a.AtPos.Row == Bau.AtRow && a.AtPos.Col == Bau.AtCol;
}

int rowAndColWithinBounds(int row, int col)
{
	return row >= 0 && row < maxRows && col >= 0 && col < maxCols;
}

void initBau()
{
	Bau.AtRow = maxRows / 2;
	Bau.AtCol = maxCols / 2;
	Bau.GatheredResources = 0;
	Bau.ResourcePosition.IsEmpty = 1;	
}



void initResource()
{	
	// determine the sugar positions
	int sugarsToCreate = maxRows * maxCols / 300; // exactly 4 sugar fields for 24 rows * 50 cols = 1200 fields
	ResourceList = mallocVector2List();
	srand (time(NULL));
	for(sugarsToCreate; sugarsToCreate > 0; sugarsToCreate--)
	{		
		int distanceFromBau;
		struct Vector2 sugarPos;
		sugarPos.IsEmpty = 0;

		do
		{
			sugarPos.Row = rand() % maxRows;
			sugarPos.Col = rand() % maxCols;
			struct Vector2 vectorToBau = getCoordinate(sugarPos.Row, sugarPos.Col, Bau.AtRow, Bau.AtCol);
			distanceFromBau = abs(vectorToBau.Row) + abs(vectorToBau.Col); 
		}
		while(distanceFromBau < 10 || (sugarPos.Row == Bau.AtRow && sugarPos.Col == Bau.AtCol));
				
		appendVector2(ResourceList, sugarPos);
	}
}



void move(struct Vector2 *curPos, struct Vector2 *movePos)
{
	if(curPos->Row + movePos->Row >= maxRows)
		movePos->Row = maxRows - curPos->Row - 1;
	if(curPos->Col + movePos->Col >= maxCols)
		movePos->Col = maxCols - curPos->Col - 1;
	if(curPos->Row + movePos->Row < 0)
		movePos->Row = -curPos->Row;
	if(curPos->Col + movePos->Col < 0)
		movePos->Col = -curPos->Col;

	int moveRow = 0, moveCol = 0;
	if( movePos->Row != 0 || movePos->Col != 0)
	{
		if(abs(movePos->Row) > abs(movePos->Col))
		{
			moveRow = abs(movePos->Row) / movePos->Row;
		}
		else
		{		
			moveCol = abs(movePos->Col) / movePos->Col;
		}
		movePos->Row -= moveRow;
		movePos->Col -= moveCol;
	}

	curPos->Row = curPos->Row+moveRow;
	curPos->Col = curPos->Col+moveCol;
}

void moveRnd(struct Vector2 *curPos, struct Vector2 *movePos)
{
	if(movePos->Row == 0 && movePos->Col == 0) // give new direction
	{
		movePos->Row = rand() % 11 - 5;
		movePos->Col = rand() % 11 - 5;
	}
	move(curPos, movePos); // do 1 step into direction
}

int antsAtPosition(struct Vector2 position)
{
	int i, count = 0; // i to iterate, count to increment the ant count at a position
	for(i = 0; i < AntsOut; i++)
	{
		if(Ants[i].AtPos.Row == position.Row && Ants[i].AtPos.Col == position.Col)
		{
			count++;
		}
	}
	return count;
}

void calculateAntMovement(struct Ant *a)
{
	if(!a->HasResource && antOnResource(*a))
	{
		a->HasResource = 1;
		a->GatheredFrom.IsEmpty = 0;
		a->GatheredFrom.Row = a->AtPos.Row;
		a->GatheredFrom.Col = a->AtPos.Col;
		giveDirectionsToBau(a);
	}
	else if(!a->HasResource && antSeesObject(a, resourceField))
	{
		struct ListV2 *headOfList = getObjectsInViewDistance(a, resourceField);
		if(headOfList->Next != NULL)
		{			
			struct Vector2 sugarCoord;
			sugarCoord = headOfList->Next->Value;
			struct Vector2 movement;
			movement = getCoordinate(a->AtPos.Row, a->AtPos.Col, sugarCoord.Row, sugarCoord.Col);		
			giveDirections(a, movement);
			move(&a->AtPos, &a->MovePos);
		}
		else
		{
			move(&a->AtPos, &a->MovePos);
		}
		freeVector2(headOfList);
	}
	else if(a->HasResource && antIsAtBau(*a))
	{
		a->HasResource = 0;
		Bau.GatheredResources++;
		if(Bau.ResourcePosition.IsEmpty)
		{
			Bau.ResourcePosition = a->GatheredFrom;			
		}
		if(a->Caste == 1) // if it's a scout
			a->GatheredFrom.IsEmpty = 1;
	}
	else if(antIsAtBau(*a) && !Bau.ResourcePosition.IsEmpty) // Bau knows about resources
	{
		if(a->GatheredFrom.IsEmpty && a->Caste == 0) // if it's a worker
		{
			a->GatheredFrom = Bau.ResourcePosition;
		}
		giveDirections(a, getCoordinate(a->AtPos.Row, a->AtPos.Col, Bau.ResourcePosition.Row, Bau.ResourcePosition.Col));
		move(&a->AtPos, &a->MovePos);
	}
	//else if(!a->HasResource && !a->GatheredFrom.IsEmpty && antsAtPosition(a->AtPos) > 1)
	//{
	//	// ants on same field share a->GatheredFrom!
	//	struct ListAnt *antList = getAntsOnSamePostion(a);
	//	struct ListAnt *currentAnt = antList;
	//	while(currentAnt->Next != NULL)
	//	{
	//		currentAnt = currentAnt->Next;
	//		if(currentAnt->Value.GatheredFrom.IsEmpty)
	//			currentAnt->Value.GatheredFrom = a->GatheredFrom;
	//	}
	//	freeAnt(antList);
	//	moveRnd(a->AtPos.Row, a->AtPos.Col, a->MoveRow, a->MoveCol);
	//}
	//else if(!a->HasResource && !a->GatheredFrom.IsEmpty && antSeesObject(a, antField))
	//{
	//	struct ListV2 *headOfList = getObjectsInViewDistance(a, antField);
	//	if(headOfList->Next != NULL)
	//	{
	//		// move to first item:
	//		struct Vector2 antCoord = headOfList->Next->Value;
	//		struct Vector2 movement = getCoordinate(a->AtPos.Row, a->AtPos.Col, antCoord.Row, antCoord.Col);
	//		movement.Row = movement.Row; // 1 step into direction
	//		movement.Col = movement.Col; // 1 step into direction
	//		giveDirections(a, movement);
	//		move(&a->AtPos, &a->MovePos);
	//	}
	//	freeVector2(headOfList);
	//}
	else
	{
		moveRnd(&a->AtPos, &a->MovePos);
	}
	
}

int main()
{	
	printf("Press any key to start");
	getchar();
	long totalTime = clock();
	long newAntTime = clock();

	/*timeval start, end;
	  gettimeofday(&start, 0);
	  ...
	  gettimeofday(&end, 0);
	*/
	consoleOutput = 1;
	fileOutput = 0;

	if(fileOutput)
	{
		f = fopen(mapFile, "r");
		if(f != NULL)
		{		
			fclose(f);
			remove(mapFile);
		}
		f = fopen(mapFile, "a+");
	}
	
	int duration, index;

	initBau();	
	initResource();
	initMap();	
	time_t start;
	time(&start);
	
	for(duration = 0; duration < SimulationLength; duration++)
	{
		int newAnt = 0;
		if(consoleOutput && (int)time(NULL) - (int)start > 0)
		{
			time(&start);
			newAnt = 1;
		}
		if(!consoleOutput)
		{
			newAnt = 1;
		}
		if(AntsOut < MaxAnts && newAnt)
		{
			int newAntCount = MaxAnts / (SimulationLength / 2);
			for(newAntCount; newAntCount > 0; newAntCount--)
			{
				addAnt(AntsOut, Bau.AtRow, Bau.AtCol);
			}
		}
		for(index = 0; index < AntsOut; index++)
		{
			calculateAntMovement(&Ants[index]);
		}
		if(consoleOutput)
		{
			system("clear");	
			printMap();			

			printf("Round: %d/%d, ", duration+1, SimulationLength);
			sleep(1);
		}
		if(fileOutput)
		{
			printMapToFile();
		}		
	}

	int timeDiff = (int)time(NULL) - (int)start;

	printf("%d Ameisen in %d seconds. Druecken Sie eine beliebige Taste zum Beenden...", AntsOut, timeDiff);
	getchar();
	
	freeVector2(ResourceList);
	return 0;
}



